@@ -41,6 +41,10 @@ public:
 		LUMINANCE_MULTIPLIER,
 		PIXEL_SIZE,
 		GLOW_INTENSITY,
+		SRGB_WHITE,
+		SSAO_INTENSITY,
+		SSAO_RADIUS_FRAC,
+		SSAO_PRN_UV,
 	};
 
 	enum ShaderVariant {
@@ -54,6 +58,11 @@ public:
 		USE_BCS = 8,
 		USE_COLOR_CORRECTION = 16,
 		USE_1D_LUT = 32,
+		USE_SSAO_ABYSS = 64,
+		USE_SSAO_LOW = 128,
+		USE_SSAO_MED = 256,
+		USE_SSAO_HIGH = 512,
+		USE_SSAO_MEGA = 1024,
 	};
 
 	_FORCE_INLINE_ bool version_bind_shader(RID p_version, ShaderVariant p_variant, uint64_t p_specialization = 0) {
@@ -226,7 +235,11 @@ protected:
 			"view",
 			"luminance_multiplier",
 			"pixel_size",
-			"glow_intensity"
+			"glow_intensity",
+			"srgb_white",
+			"ssao_intensity",
+			"ssao_radius_frac",
+			"ssao_prn_UV"
 		};
 		static const char *_variant_defines[] = {
 			"",
@@ -235,6 +248,8 @@ protected:
 			{ "source_color", 0 },
 			{ "glow_color", 1 },
 			{ "source_color_correction", 2 },
+			{ "depth_buffer_array", 3 },
+			{ "depth_buffer", 3 },
 		};
 		static UBOPair _ubo_pairs[] = {
 			{ "TonemapData", 0 },
@@ -246,6 +261,11 @@ protected:
 			{ "USE_BCS", false },
 			{ "USE_COLOR_CORRECTION", false },
 			{ "USE_1D_LUT", false },
+			{ "USE_SSAO_ABYSS", false },
+			{ "USE_SSAO_LOW", false },
+			{ "USE_SSAO_MED", false },
+			{ "USE_SSAO_HIGH", false },
+			{ "USE_SSAO_MEGA", false },
 		};
 		static const Feedback *_feedbacks = nullptr;
 		static const char _vertex_code[] = {
@@ -265,21 +285,21 @@ void main() {
 		};
 
 		static const char _fragment_code[] = {
-R"<!>(/* clang-format on */
+(R"<!>(/* clang-format on */
 
 // If we reach this code, we always tonemap.
 #define APPLY_TONEMAPPING
 
 layout(std140) uniform TonemapData { //ubo:0
 	float exposure;
-	float white;
 	int tonemapper;
 	int pad;
-
 	int pad2;
+	vec4 tonemapper_params;
 	float brightness;
 	float contrast;
 	float saturation;
+	int pad3;
 };
 
 // This expects 0-1 range input.
@@ -300,17 +320,18 @@ vec3 srgb_to_linear(vec3 color) {
 #ifdef APPLY_TONEMAPPING
 
 // Based on Reinhard's extended formula, see equation 4 in https://doi.org/cjbgrt
-vec3 tonemap_reinhard(vec3 color, float p_white) {
-	float white_squared = p_white * p_white;
+vec3 tonemap_reinhard(vec3 color) {
+	float white_squared = tonemapper_params.x;
 	vec3 white_squared_color = white_squared * color;
 	// Equivalent to color * (1 + color / white_squared) / (1 + color)
 	return (white_squared_color + color * color) / (white_squared_color + white_squared);
 }
 
-vec3 tonemap_filmic(vec3 color, float p_white) {
-	// exposure bias: input scale (color *= bias, white *= bias) to make the brightness consistent with other tonemappers
-	// also useful to scale the input to the range that the tonemapper is designed for (some require very high input values)
-	// has no effect on the curve's general shape or visual properties
+vec3 tonemap_filmic(vec3 color) {
+	// These constants must match the those in the C++ code that calculates the parameters.
+	// exposure_bias: Input scale (color *= bias, env->white *= bias) to make the brightness consistent with other tonemappers.
+	// Also useful to scale the input to the range that the tonemapper is designed for (some require very high input values).
+	// Has no effect on the curve's general shape or visual properties.
 	const float exposure_bias = 2.0f;
 	const float A = 0.22f * exposure_bias * exposure_bias; // bias baked into constants for performance
 	const float B = 0.30f * exposure_bias;
@@ -320,14 +341,14 @@ vec3 tonemap_filmic(vec3 color, float p_white) {
 	const float F = 0.30f;
 
 	vec3 color_tonemapped = ((color * (A * color + C * B) + D * E) / (color * (A * color + B) + D * F)) - E / F;
-	float p_white_tonemapped = ((p_white * (A * p_white + C * B) + D * E) / (p_white * (A * p_white + B) + D * F)) - E / F;
 
-	return color_tonemapped / p_white_tonemapped;
+	return color_tonemapped / tonemapper_params.x;
 }
 
 // Adapted from https://github.com/TheRealMJP/BakingLab/blob/master/BakingLab/ACES.hlsl
 // (MIT License).
-vec3 tonemap_aces(vec3 color, float p_white) {
+vec3 tonemap_aces(vec3 color) {
+	// These constants must match the those in the C++ code that calculates the parameters.
 	const float exposure_bias = 1.8f;
 	const float A = 0.0245786f;
 	const float B = 0.000090537f;
@@ -350,46 +371,46 @@ vec3 tonemap_aces(vec3 color, float p_white) {
 	vec3 color_tonemapped = (color * (color + A) - B) / (color * (C * color + D) + E);
 	color_tonemapped *= odt_to_rgb;
 
-	p_white *= exposure_bias;
-	float p_white_tonemapped = (p_white * (p_white + A) - B) / (p_white * (C * p_white + D) + E);
-
-	return color_tonemapped / p_white_tonemapped;
+	return color_tonemapped / tonemapper_params.x;
 }
 
-// Polynomial approximation of EaryChow's AgX sigmoid curve.
-// x must be within the range [0.0, 1.0]
-vec3 agx_contrast_approx(vec3 x) {
-	// Generated with Excel trendline
-	// Input data: Generated using python sigmoid with EaryChow's configuration and 57 steps
-	// Additional padding values were added to give correct intersections at 0.0 and 1.0
-	// 6th order, intercept of 0.0 to remove an operation and ensure intersection at 0.0
-	vec3 x2 = x * x;
-	vec3 x4 = x2 * x2;
-	return 0.021 * x + 4.0111 * x2 - 25.682 * x2 * x + 70.359 * x4 - 74.778 * x4 * x + 27.069 * x4 * x2;
+// allenwp tonemapping curve; developed for use in the Godot game engine.
+// Source and details: https://allenwp.com/blog/2025/05/29/allenwp-tonemapping-curve/
+// Input must be a non-negative linear scene value.
+vec3 allenwp_curve(vec3 x) {
+	const float output_max_value = 1.0; // SDR always has an output_max_value of 1.0
+
+	// These constants must match the those in the C++ code that calculates the parameters.
+	// 18% "middle gray" is perceptually 50% of the brightness of reference white.
+	const float awp_crossover_point = 0.18;
+	// When output_max_value and/or awp_crossover_point are no longer constant,
+	// awp_shoulder_max can be calculated on the CPU and passed in as tonemap_e.
+	const float awp_shoulder_max = output_max_value - awp_crossover_point;
+
+	float awp_contrast = tonemapper_params.x;
+	float awp_toe_a = tonemapper_params.y;
+	float awp_slope = tonemapper_params.z;
+	float awp_w = tonemapper_params.w;
+
+	// Reinhard-like shoulder:
+	vec3 s = x - awp_crossover_point;
+	vec3 slope_s = awp_slope * s;
+	s = slope_s * (1.0 + s / awp_w) / (1.0 + (slope_s / awp_shoulder_max));
+	s += awp_crossover_point;
+
+	// Sigmoid power function toe:
+	vec3 t = pow(x, vec3(awp_contrast));
+	t = t / (t + awp_toe_a);
+
+	return mix(s, t, lessThan(x, vec3(awp_crossover_point)));
 }
 
 // This is an approximation and simplification of EaryChow's AgX implementation that is used by Blender.
 // This code is based off of the script that generates the AgX_Base_sRGB.cube LUT that Blender uses.
 // Source: https://github.com/EaryChow/AgX_LUT_Gen/blob/main/AgXBasesRGB.py
+// Colorspace transformation source: https://www.colour-science.org:8010/apps/rgb_colourspace_transformation_matrix
 vec3 tonemap_agx(vec3 color) {
-	// Combined linear sRGB to linear Rec 2020 and Blender AgX inset matrices:
-	const mat3 srgb_to_rec2020_agx_inset_matrix = mat3(
-			0.54490813676363087053, 0.14044005884001287035, 0.088827411851915368603,
-			0.37377945959812267119, 0.75410959864013760045, 0.17887712465043811023,
-			0.081384976686407536266, 0.10543358536857773485, 0.73224999956948382528);
-
-	// Combined inverse AgX outset matrix and linear Rec 2020 to linear sRGB matrices.
-	const mat3 agx_outset_rec2020_to_srgb_matrix = mat3(
-			1.9645509602733325934, -0.29932243390911083839, -0.16436833806080403409,
-			-0.85585845117807513559, 1.3264510741502356555, -0.23822464068860595117,
-			-0.10886710826831608324, -0.027084020983874825605, 1.402665347143271889);
-
-	// LOG2_MIN      = -10.0
-	// LOG2_MAX      =  +6.5
-	// MIDDLE_GRAY   =  0.18
-	const float min_ev = -12.4739311883324; // log2(pow(2, LOG2_MIN) * MIDDLE_GRAY)
-	const float max_ev = 4.02606881166759; // log2(pow(2, LOG2_MAX) * MIDDLE_GRAY)
-
+	// Input color should be non-negative!
 	// Large negative values in one channel and large positive values in other
 	// channels can result in a colour that appears darker and more saturated than
 	// desired after passing it through the inset matrix. For this reason, it is
@@ -397,31 +418,43 @@ vec3 tonemap_agx(vec3 color) {
 	// This is done before the Rec. 2020 transform to allow the Rec. 2020
 	// transform to be combined with the AgX inset matrix. This results in a loss
 	// of color information that could be correctly interpreted within the
-	// Rec. 2020 color space as positive RGB values, but it is less common for Godot
-	// to provide this function with negative sRGB values and therefore not worth
+	// Rec. 2020 color space as positive RGB values, but is often not worth
 	// the performance cost of an additional matrix multiplication.
-	// A value of 2e-10 intentionally introduces insignificant error to prevent
-	// log2(0.0) after the inset matrix is applied; color will be >= 1e-10 after
-	// the matrix transform.
-	color = max(color, 2e-10);
+	//
+	// Additionally, this AgX configuration was created subjectively based on
+	// output appearance in the Rec. 709 color gamut, so it is possible that these
+	// matrices will not perform well with non-Rec. 709 output (more testing with
+	// future wide-gamut displays is be needed).
+	// See this comment from the author on the decisions made to create the matrices:
+	// https://github.com/godotengine/godot-proposals/issues/12317#issuecomment-2835824250
 
-	// Do AGX in rec2020 to match Blender and then apply inset matrix.
-	color = srgb_to_rec2020_agx_inset_matrix * color;
+	// Combined Rec. 709 to Rec. 2020 and Blender AgX inset matrices:
+	const mat3 rec709_to_rec2020_agx_inset_matrix = mat3(
+			0.544814746488245, 0.140416948464053, 0.0888104196149096,
+			0.373787398372697, 0.754137554567394, 0.178871756420858,
+			0.0813978551390581, 0.105445496968552, 0.732317823964232);
 
-	// Log2 space encoding.
-	// Must be clamped because agx_contrast_approx may not work
-	// well with values outside of the range [0.0, 1.0]
-	color = clamp(log2(color), min_ev, max_ev);
-	color = (color - min_ev) / (max_ev - min_ev);
+	// Combined inverse AgX outset matrix and Rec. 2020 to Rec. 709 matrices.
+	const mat3 agx_outset_rec2020_to_rec709_matrix = mat3(
+			1.96488741169489, -0.299313364904742, -0.164352742528393,
+			-0.855988495690215, 1.32639796461980, -0.238183969428088,
+			-0.108898916004672, -0.0270845997150571, 1.40253671195648);
 
-	// Apply sigmoid function approximation.
-	color = agx_contrast_approx(color);
+	const float output_max_value = 1.0; // SDR always has an output_max_value of 1.0
 
-	// Convert back to linear before applying outset matrix.
-	color = pow(color, vec3(2.4));
+	// Apply inset matrix.
+	color = rec709_to_rec2020_agx_inset_matrix * color;
 
-	// Apply outset to make the result more chroma-laden and then go back to linear sRGB.
-	color = agx_outset_rec2020_to_srgb_matrix * color;
+	// Use the allenwp tonemapping curve to match the Blender AgX curve while
+	// providing stability across all variable dyanimc range (SDR, HDR, EDR).
+	color = allenwp_curve(color);
+
+	// Clipping to output_max_value is required to address a cyan colour that occurs
+	// with very bright inputs.
+	color = min(vec3(output_max_value), color);
+
+	// Apply outset to make the result more chroma-laden and then go back to Rec. 709.
+	color = agx_outset_rec2020_to_rec709_matrix * color;
 
 	// Blender's lusRGB.compensate_low_side is too complex for this shader, so
 	// simply return the color, even if it has negative components. These negative
@@ -435,17 +468,21 @@ vec3 tonemap_agx(vec3 color) {
 #define TONEMAPPER_ACES 3
 #define TONEMAPPER_AGX 4
 
-vec3 apply_tonemapping(vec3 color, float p_white) { // inputs are LINEAR
-	// Ensure color values passed to tonemappers are positive.
-	// They can be negative in the case of negative lights, which leads to undesired behavior.
+vec3 apply_tonemapping(vec3 color) { // inputs are LINEAR
 	if (tonemapper == TONEMAPPER_LINEAR) {
 		return color;
-	} else if (tonemapper == TONEMAPPER_REINHARD) {
-		return tonemap_reinhard(max(vec3(0.0f), color), p_white);
+	}
+
+	// Ensure color values passed to tonemappers are positive.
+	// They can be negative in the case of negative lights, which leads to undesired behavior.
+	color = max(vec3(0.0), color);
+
+	if (tonemapper == TONEMAPPER_REINHARD) {
+		return tonemap_reinhard(color);
 	} else if (tonemapper == TONEMAPPER_FILMIC) {
-		return tonemap_filmic(max(vec3(0.0f), color), p_white);
+		return tonemap_filmic(color);
 	} else if (tonemapper == TONEMAPPER_ACES) {
-		return tonemap_aces(max(vec3(0.0f), color), p_white);
+		return tonemap_aces(color);
 	} else { // TONEMAPPER_AGX
 		return tonemap_agx(color);
 	}
@@ -466,6 +503,7 @@ uniform float luminance_multiplier;
 uniform sampler2D glow_color; // texunit:1
 uniform vec2 pixel_size;
 uniform float glow_intensity;
+uniform float srgb_white;
 
 vec4 get_glow_color(vec2 uv) {
 	vec2 half_pixel = pixel_size * 0.5;
@@ -479,6 +517,10 @@ vec4 get_glow_color(vec2 uv) {
 	color += textureLod(glow_color, uv + vec2(0.0, -half_pixel.y * 2.0), 0.0);
 	color += textureLod(glow_color, uv + vec2(-half_pixel.x, -half_pixel.y), 0.0) * 2.0;
 
+#ifdef USE_LUMINANCE_MULTIPLIER
+	color = color / luminance_multiplier;
+#endif
+
 	return color / 12.0;
 }
 #endif // USE_GLOW
@@ -502,14 +544,169 @@ vec3 apply_color_correction(vec3 color) {
 #endif // USE_1D_LUT
 #endif // USE_COLOR_CORRECTION
 
-#ifdef USE_BCS
-vec3 apply_bcs(vec3 color) {
-	color = mix(vec3(0.0), color, brightness);
-	color = mix(vec3(0.5), color, contrast);
-	color = mix(vec3(dot(vec3(1.0), color) * 0.33333), color, saturation);
+#if defined(USE_SSAO_ABYSS) || defined(USE_SSAO_LOW) || defined(USE_SSAO_MED) || defined(USE_SSAO_HIGH) || defined(USE_SSAO_MEGA)
+#define USE_SOME_SSAO
+uniform float ssao_intensity;
+uniform float ssao_radius_frac;
+uniform vec2 ssao_prn_UV;
+#ifdef USE_MULTIVIEW
+// VR will have 2 depth buffers.
+uniform sampler2DArray depth_buffer_array; // texunit:3
+#else
+uniform sampler2D depth_buffer; // texunit:3
+#endif
+#if defined(USE_SSAO_ABYSS)
+// Use the tiny 2-sample version.
+// S4AO (Stupid Simple Screen Space Ambient Occlusion) - Jonathan Dummer (O1S)
+// This micro version uses only 3 depth samples, the midpoint and a randomly-rotated, balanced pair.
 
-	return color;
+const mediump float ssao_falloff_frac = 0.25;
+// Perform the SSAO.
+float s4ao(vec2 UV) {
+#ifdef USE_MULTIVIEW
+	mediump float depth = texture(depth_buffer_array, vec3(UV, view)).r;
+#else
+	mediump float depth = texture(depth_buffer, UV).r;
+#endif
+	mediump float inv_falloff = 1.0f / max(1e-4f, depth * ssao_falloff_frac);
+	// Random 2D rotation per pixel (0..1 -> parabola approximating a 180 deg arc)
+	mediump float r01 = fract(dot(UV, ssao_prn_UV));
+	mediump vec2 duv = vec2(r01 - 0.5f, 2.0f * (r01 - r01 * r01)) * (2.0f * depth * ssao_radius_frac); // 180 degrees.
+	// Grab the samples and determine the occlusion.
+	mediump float occlusion = 0.0f;
+	for (int s = 0; s < 2; ++s) {
+#ifdef USE_MULTIVIEW
+		mediump float dz = texture(depth_buffer_array, vec3(UV + duv, view)).r - depth;
+#else
+		mediump float dz = texture(depth_buffer, UV + duv).r - depth;
+#endif
+		// How 'directly overhead' is it?  Factor in the falloff depth.
+		occlusion += normalize(vec3(duv, dz)).z * mix(1.0f, 0.0f, dz * inv_falloff);
+		// Mirror the next sample.
+		duv = -duv;
+	}
+	// Adjust the occlusion for intensity, and # samples.
+	occlusion = 1.0f - clamp(occlusion * 0.5f * ssao_intensity, 0.0f, 1.0f);
+	return occlusion * occlusion;
 }
+#elif defined(USE_SSAO_HIGH) || defined(USE_SSAO_MEGA)
+// Use the rings version for the higher qualities.
+// S4AO (Stupid Simple Screen Space Ambient Occlusion) - Jonathan Dummer (O1S)
+// The mega version uses N concentric rings of samples.
+
+#if defined(USE_SSAO_MEGA)
+const int rings = 4; // Start with the outer ring.
+const int samps[] = int[](24, 18, 12, 6); // ( 9, 6, 3 ) is a minimum, but I want better.
+#else
+const int rings = 3; // Start with the outer ring.
+const int samps[] = int[](15, 10, 5, 1); // ( 9, 6, 3 ) is a minimum, but I want better.
+#endif
+const float average_samples = 1.0 / float(samps[0] + samps[1] * int(rings > 1) + samps[2] * int(rings > 2) + samps[3] * int(rings > 3));
+const float ssao_falloff_frac = 0.25;
+// Perform the SSAO.
+float s4ao(vec2 UV) {
+#ifdef USE_MULTIVIEW
+	float depth = texture(depth_buffer_array, vec3(UV, view)).r;
+#else
+	float depth = texture(depth_buffer, UV).r;
+#endif
+	float inv_falloff = 1.0f / max(1e-4f, depth * ssao_falloff_frac);
+	// Random 2D rotation per pixel (0..1 -> parabola approximating a 180 deg arc)
+	float r01 = fract(dot(UV, ssao_prn_UV));
+	vec2 rcos = vec2(r01 - 0.5f, 2.0f * (r01 - r01 * r01)) * (2.0f * depth * ssao_radius_frac); // 180 degrees.
+	vec2 rsin = rcos.yx * vec2(-1, 1); // Perpendicular to the random cosine vector.
+	// Grab the samples and determine the occlusion.
+	float occlusion = 0.0f;
+	float ring_shrink = 0.75f; // Shrink every ring.
+	for (int r = 0; r < rings; ++r) {
+		float dt = (6.283185307f) / float(samps[r]);
+		float t = float(r & 1) * 0.5f * dt;
+		for (int s = 0; s < samps[r]; ++s) {
+			vec2 duv = cos(t) * rcos + sin(t) * rsin;
+#ifdef USE_MULTIVIEW
+			float dz = texture(depth_buffer_array, vec3(UV + duv, view)).r - depth;
+#else
+			float dz = texture(depth_buffer, UV + duv).r - depth;
+#endif
+			// How 'directly overhead' is it?  Factor in the falloff depth.
+			occlusion += normalize(vec3(duv, dz)).z * smoothstep(1.0f, 0.0f, dz * inv_falloff);
+			t += dt;
+		}
+		// The next ring will be smaller.
+		rcos *= ring_shrink;
+		rsin *= ring_shrink;
+	}
+	// Adjust the occlusion for intensity, and # samples.
+	occlusion *= ssao_intensity * average_samples;
+	occlusion = 1.0f - clamp(occlusion, 0.0f, 1.0f);
+	return occlusion * occlusion;
+}
+#else
+// Use the more generic NxN grid version.
+// S4AO (Stupid Simple Screen Space Ambient Occlusion) - Jonathan Dummer (O1S)
+)<!>" R"<!>(
+// The sample_width should be even, else the midpoint is at UV.
+// Takes sample_width^2 samples in a grid, with the corners notched.
+#if defined(USE_SSAO_LOW)
+const int sample_width = 2;
+#elif defined(USE_SSAO_HIGH)
+const int sample_width = 6;
+#else
+const int sample_width = 4;
+#endif
+const int notch_01 = int(sample_width > 3); // Set to 1 to skip the corner samples, 0 to include them.
+const float sample_mid = (float(sample_width) - 1.0) * 0.50001; // Can't be exactly 0.5 in case sample_width is odd.
+#if defined(USE_SSAO_LOW)
+const float inv_half_width = 1.0 / sample_mid; // The 2x2 sampling looks wider as all samples are at radius.
+#else
+const float inv_half_width = 1.7 / sample_mid; // Bake in the 1.7 scale for the random rotation.
+#endif
+const float average_samples = 1.0 / float(sample_width * sample_width - 4 * notch_01); //  1 / number_of_samples
+const float ssao_falloff_frac = 0.25;
+// Perform the SSAO.
+float s4ao(vec2 UV) {
+#ifdef USE_MULTIVIEW
+	float depth = texture(depth_buffer_array, vec3(UV, view)).r;
+#else
+	float depth = texture(depth_buffer, UV).r;
+#endif
+	float radius = max(1e-4f, depth * ssao_radius_frac);
+	float inv_falloff = 1.0f / max(1e-4f, depth * ssao_falloff_frac);
+	// Random 2D rotation per pixel (+/-45 deg, with 0 having a lower probability).
+	// The random cosine vector is vec2( 0.5, -0.5 to +0.5 ) and *1.7 makes the average length ~ 1.
+	vec2 rcos = (inv_half_width * radius) * vec2(0.5f, fract(dot(UV, ssao_prn_UV)) - 0.5f);
+	vec2 rsin = rcos.yx * vec2(-1, 1); // Perpendicular to the random cosine vector.
+	// Grab the samples and determine the occlusion.
+	float occlusion = 0.0f;
+	vec2 base_duv = -sample_mid * rsin;
+	for (int j = sample_width; --j >= 0;) {
+#if defined(USE_SSAO_LOW)
+		// Low quality uses 2x2 samples, no notching.
+		vec2 duv = -sample_mid * rcos + base_duv;
+		for (int i = sample_width; --i >= 0;) {
+#else
+		//	Will uses 4x4 or 6x6 samples, with the corners notched out.
+		int o = /*notch_01 &*/ int((j <= 0) || (j >= (sample_width - 1))); // Notch corners of the grid.
+		vec2 duv = (float(o) - sample_mid) * rcos + base_duv;
+		for (int i = sample_width - o - o; --i >= 0;) {
+#endif
+#ifdef USE_MULTIVIEW
+			float dz = texture(depth_buffer_array, vec3(UV + duv, view)).r - depth;
+#else
+			float dz = texture(depth_buffer, UV + duv).r - depth;
+#endif
+			float validity = smoothstep(1.0f, 0.0f, dz * inv_falloff);
+			occlusion += normalize(vec3(duv, dz)).z * validity; // How 'directly overhead' is it?
+			duv += rcos; // March along the rcos direction with i.
+		}
+		base_duv += rsin; // March along the rsin direction with j.
+	}
+	// Adjust the occlusion for intensity, and # samples.
+	occlusion *= ssao_intensity * average_samples;
+	occlusion = clamp(1.0f - occlusion, 0.0f, 1.0f);
+	return occlusion * occlusion;
+}
+#endif
 #endif
 
 in vec2 uv_interp;
@@ -523,25 +720,71 @@ void main() {
 	vec4 color = texture(source_color, uv_interp);
 #endif
 
-#ifdef USE_GLOW
-	vec4 glow = get_glow_color(uv_interp) * glow_intensity;
-
-	// Just use softlight...
-	glow.rgb = clamp(glow.rgb, vec3(0.0f), vec3(1.0f));
-	color.rgb = max((color.rgb + glow.rgb) - (color.rgb * glow.rgb), vec3(0.0));
-#endif // USE_GLOW
-
 #ifdef USE_LUMINANCE_MULTIPLIER
 	color = color / luminance_multiplier;
 #endif
 
+#ifdef USE_GLOW
+	// Glow blending is performed before srgb_to_linear because
+	// the glow texture was created from a nonlinear sRGB-encoded
+	// scene, so it only makes sense to add this glow to an equally
+	// nonlinear sRGB-encoded scene.
+
+	vec4 glow = get_glow_color(uv_interp) * glow_intensity;
+
+	// Glow always uses the screen blend mode in the Compatibility renderer:
+
+	// Glow cannot be above 1.0 after normalizing and should be non-negative
+	// to produce expected results. It is possible that glow can be negative
+	// if negative lights were used in the scene.
+	// We clamp to srgb_white because glow will be normalized to this range.
+	// Note: srgb_white cannot be smaller than the maximum output value (1.0).
+	glow.rgb = clamp(glow.rgb, 0.0, srgb_white);
+
+	// Normalize to srgb_white range.
+	//glow.rgb /= srgb_white;
+	//color.rgb /= srgb_white;
+	//color.rgb = (color.rgb + glow.rgb) - (color.rgb * glow.rgb);
+	// Expand back to original range.
+	//color.rgb *= srgb_white;
+
+	// The following is a mathematically simplified version of the above.
+	color.rgb = color.rgb + glow.rgb - (color.rgb * glow.rgb / srgb_white);
+#endif // USE_GLOW
+
 	color.rgb = srgb_to_linear(color.rgb);
-	color.rgb = apply_tonemapping(color.rgb, white);
-	color.rgb = linear_to_srgb(color.rgb);
+
+#if defined(USE_SOME_SSAO)
+	// Putting SSAO after the conversion to linear color, though it might be better before the glow.
+	color.rgb *= s4ao(uv_interp); // The USE_SSAO_X controls the number of samples.
+#endif
+
+	color.rgb = apply_tonemapping(color.rgb);
 
 #ifdef USE_BCS
-	color.rgb = apply_bcs(color.rgb);
-#endif
+	// Apply brightness:
+	// Apply to relative luminance. This ensures that the hue and saturation of
+	// colors is not affected by the adjustment, but requires the multiplication
+	// to be performed on linear-encoded values.
+	color.rgb = color.rgb * brightness;
+
+	color.rgb = linear_to_srgb(color.rgb);
+
+	// Apply contrast:
+	// By applying contrast to RGB values that are perceptually uniform (nonlinear),
+	// the darkest values are not hard-clipped as badly, which produces a
+	// higher quality contrast adjustment and maintains compatibility with
+	// existing projects.
+	color.rgb = mix(vec3(0.5), color.rgb, contrast);
+
+	// Apply saturation:
+	// By applying saturation adjustment to nonlinear sRGB-encoded values with
+	// even weights the preceived brightness of blues are affected, but this
+	// maintains compatibility with existing projects.
+	color.rgb = mix(vec3(dot(vec3(1.0), color.rgb) * (1.0 / 3.0)), color.rgb, saturation);
+#else
+	color.rgb = linear_to_srgb(color.rgb);
+#endif // USE_BCS
 
 #ifdef USE_COLOR_CORRECTION
 	color.rgb = apply_color_correction(color.rgb);
@@ -549,12 +792,12 @@ void main() {
 
 	frag_color = color;
 }
-)<!>"
+)<!>")
 		};
 
 		_setup(_vertex_code, _fragment_code, "PostShaderGLES3",
-				4, _uniform_strings, 1, _ubo_pairs,
-				0, _feedbacks, 3, _texunit_pairs,
-				6, _spec_pairs, 1, _variant_defines);
+				8, _uniform_strings, 1, _ubo_pairs,
+				0, _feedbacks, 5, _texunit_pairs,
+				11, _spec_pairs, 1, _variant_defines);
 	}
 };
