@@ -32,7 +32,7 @@
 
 #include "scene/gui/container.h"
 #include "scene/theme/theme_db.h"
-#include "servers/text_server.h"
+#include "servers/text/text_server.h"
 
 void Label::set_autowrap_mode(TextServer::AutowrapMode p_mode) {
 	if (autowrap_mode == p_mode) {
@@ -136,6 +136,8 @@ int Label::get_line_height(int p_line) const {
 }
 
 void Label::_shape() const {
+	const String &lang = language.is_empty() ? _get_locale() : language;
+
 	Ref<StyleBox> style = theme_cache.normal_style;
 	int width = (get_size().width - style->get_minimum_size().width);
 
@@ -149,7 +151,7 @@ void Label::_shape() const {
 		}
 		paragraphs.clear();
 
-		String txt = (uppercase) ? TS->string_to_upper(xl_text, language) : xl_text;
+		String txt = (uppercase) ? TS->string_to_upper(xl_text, lang) : xl_text;
 		if (visible_chars >= 0 && visible_chars_behavior == TextServer::VC_CHARS_BEFORE_SHAPING) {
 			txt = txt.substr(0, visible_chars);
 		}
@@ -183,7 +185,7 @@ void Label::_shape() const {
 			ERR_FAIL_COND(font.is_null());
 
 			if (para.dirty) {
-				TS->shaped_text_add_string(para.text_rid, para.text, font->get_rids(), font_size, font->get_opentype_features(), language);
+				TS->shaped_text_add_string(para.text_rid, para.text, font->get_rids(), font_size, font->get_opentype_features(), lang);
 			} else {
 				int spans = TS->shaped_get_span_count(para.text_rid);
 				for (int i = 0; i < spans; i++) {
@@ -256,38 +258,7 @@ void Label::_shape() const {
 		}
 
 		if (para.lines_dirty) {
-			BitField<TextServer::TextOverrunFlag> overrun_flags = TextServer::OVERRUN_NO_TRIM;
-			switch (overrun_behavior) {
-				case TextServer::OVERRUN_TRIM_WORD_ELLIPSIS_FORCE: {
-					overrun_flags.set_flag(TextServer::OVERRUN_TRIM);
-					overrun_flags.set_flag(TextServer::OVERRUN_TRIM_WORD_ONLY);
-					overrun_flags.set_flag(TextServer::OVERRUN_ADD_ELLIPSIS);
-					overrun_flags.set_flag(TextServer::OVERRUN_ENFORCE_ELLIPSIS);
-				} break;
-				case TextServer::OVERRUN_TRIM_ELLIPSIS_FORCE: {
-					overrun_flags.set_flag(TextServer::OVERRUN_TRIM);
-					overrun_flags.set_flag(TextServer::OVERRUN_ADD_ELLIPSIS);
-					overrun_flags.set_flag(TextServer::OVERRUN_ENFORCE_ELLIPSIS);
-				} break;
-				case TextServer::OVERRUN_TRIM_WORD_ELLIPSIS:
-					overrun_flags.set_flag(TextServer::OVERRUN_TRIM);
-					overrun_flags.set_flag(TextServer::OVERRUN_TRIM_WORD_ONLY);
-					overrun_flags.set_flag(TextServer::OVERRUN_ADD_ELLIPSIS);
-					break;
-				case TextServer::OVERRUN_TRIM_ELLIPSIS:
-					overrun_flags.set_flag(TextServer::OVERRUN_TRIM);
-					overrun_flags.set_flag(TextServer::OVERRUN_ADD_ELLIPSIS);
-					break;
-				case TextServer::OVERRUN_TRIM_WORD:
-					overrun_flags.set_flag(TextServer::OVERRUN_TRIM);
-					overrun_flags.set_flag(TextServer::OVERRUN_TRIM_WORD_ONLY);
-					break;
-				case TextServer::OVERRUN_TRIM_CHAR:
-					overrun_flags.set_flag(TextServer::OVERRUN_TRIM);
-					break;
-				case TextServer::OVERRUN_NO_TRIMMING:
-					break;
-			}
+			BitField<TextServer::TextOverrunFlag> overrun_flags = TextServer::get_overrun_flags_from_behavior(overrun_behavior);
 
 			// Fill after min_size calculation.
 
@@ -688,6 +659,43 @@ PackedStringArray Label::get_configuration_warnings() const {
 		}
 	}
 
+	Ref<FontFile> ff = font;
+	if (ff.is_valid() && ff->is_multichannel_signed_distance_field()) {
+		bool has_settings = settings.is_valid();
+		int font_size = settings.is_valid() ? settings->get_font_size() : theme_cache.font_size;
+		int outline_size = has_settings ? settings->get_outline_size() : theme_cache.font_outline_size;
+		Vector<LabelSettings::StackedOutlineData> stacked_outline_datas = has_settings ? settings->get_stacked_outline_data() : Vector<LabelSettings::StackedOutlineData>();
+		Vector<LabelSettings::StackedShadowData> stacked_shadow_datas = has_settings ? settings->get_stacked_shadow_data() : Vector<LabelSettings::StackedShadowData>();
+		int max_outline_draw_size = outline_size;
+		if (stacked_outline_datas.size() != 0) {
+			int draw_iterations = stacked_outline_datas.size();
+			for (int j = 0; j < draw_iterations; j++) {
+				int stacked_outline_size = stacked_outline_datas[j].size;
+				if (stacked_outline_size <= 0) {
+					continue;
+				}
+				max_outline_draw_size += stacked_outline_size;
+			}
+		}
+		if (stacked_shadow_datas.size() != 0) {
+			int draw_iterations = stacked_shadow_datas.size();
+			for (int j = 0; j < draw_iterations; j++) {
+				LabelSettings::StackedShadowData stacked_shadow_data = stacked_shadow_datas[j];
+				if (stacked_shadow_data.outline_size > 0) {
+					max_outline_draw_size = MAX(max_outline_draw_size, stacked_shadow_data.outline_size);
+				}
+			}
+		}
+		float scale = (float)font_size / (float)ff->get_msdf_size();
+		float ol = (float)max_outline_draw_size / scale / 4.0;
+		float pxr = (float)ff->get_msdf_pixel_range() / 2.0 - 1.0;
+		float r_pxr = (ol + 1.0) * 2.0;
+
+		if (ol > pxr) {
+			warnings.push_back(vformat(RTR("MSDF font pixel range is too small, some outlines/shadows will not render. Set MSDF pixel range to be at least %d to render all outlines/shadows."), Math::ceil(r_pxr)));
+		}
+	}
+
 	return warnings;
 }
 
@@ -703,15 +711,14 @@ void Label::_notification(int p_what) {
 		} break;
 
 		case NOTIFICATION_TRANSLATION_CHANGED: {
-			String new_text = atr(text);
-			if (new_text == xl_text) {
-				return; // Nothing new.
+			const String new_text = atr(text);
+			if (new_text != xl_text) {
+				xl_text = new_text;
+				if (visible_ratio < 1) {
+					visible_chars = get_total_character_count() * visible_ratio;
+				}
 			}
-			xl_text = new_text;
-			if (visible_ratio < 1) {
-				visible_chars = get_total_character_count() * visible_ratio;
-			}
-			text_dirty = true;
+			text_dirty = true; // Language update might change the appearance of some characters.
 
 			queue_accessibility_update();
 			queue_redraw();
@@ -762,7 +769,7 @@ void Label::_notification(int p_what) {
 			Vector<LabelSettings::StackedShadowData> stacked_shadow_datas = has_settings ? settings->get_stacked_shadow_data() : Vector<LabelSettings::StackedShadowData>();
 			bool rtl_layout = is_layout_rtl();
 
-			if (has_focus()) {
+			if (has_focus(true)) {
 				theme_cache.focus_style->draw(ci, Rect2(Point2(0, 0), get_size()));
 			} else {
 				theme_cache.normal_style->draw(ci, Rect2(Point2(0, 0), get_size()));
@@ -888,6 +895,7 @@ void Label::_notification(int p_what) {
 		case NOTIFICATION_THEME_CHANGED: {
 			font_dirty = true;
 			queue_redraw();
+			update_configuration_warnings();
 		} break;
 
 		case NOTIFICATION_RESIZED: {
@@ -974,7 +982,13 @@ Size2 Label::get_minimum_size() const {
 
 	Size2 min_style = theme_cache.normal_style->get_minimum_size();
 	if (autowrap_mode != TextServer::AUTOWRAP_OFF) {
-		return Size2(1, (clip || overrun_behavior != TextServer::OVERRUN_NO_TRIMMING) ? 1 : min_size.height) + min_style;
+		if (!clip && overrun_behavior != TextServer::OVERRUN_NO_TRIMMING && max_lines_visible > 0) {
+			int line_spacing = settings.is_valid() ? settings->get_line_spacing() : theme_cache.line_spacing;
+			min_size.height = MIN(min_size.height, (font->get_height(font_size) + line_spacing) * max_lines_visible);
+		} else if (clip || overrun_behavior != TextServer::OVERRUN_NO_TRIMMING) {
+			min_size.height = 1;
+		}
+		return Size2(1, min_size.height) + min_style;
 	} else {
 		if (clip || overrun_behavior != TextServer::OVERRUN_NO_TRIMMING) {
 			min_size.width = 1;
@@ -1101,6 +1115,7 @@ void Label::set_text(const String &p_string) {
 void Label::_invalidate() {
 	font_dirty = true;
 	queue_redraw();
+	update_configuration_warnings();
 }
 
 void Label::set_label_settings(const Ref<LabelSettings> &p_settings) {
@@ -1113,6 +1128,7 @@ void Label::set_label_settings(const Ref<LabelSettings> &p_settings) {
 			settings->connect_changed(callable_mp(this, &Label::_invalidate), CONNECT_REFERENCE_COUNTED);
 		}
 		_invalidate();
+		update_configuration_warnings();
 	}
 }
 
@@ -1145,12 +1161,12 @@ TextServer::StructuredTextParser Label::get_structured_text_bidi_override() cons
 	return st_parser;
 }
 
-void Label::set_structured_text_bidi_override_options(Array p_args) {
+void Label::set_structured_text_bidi_override_options(const Array &p_args) {
 	if (st_args == p_args) {
 		return;
 	}
 
-	st_args = p_args;
+	st_args = Array(p_args);
 	for (Paragraph &para : paragraphs) {
 		para.dirty = true;
 	}
@@ -1158,7 +1174,7 @@ void Label::set_structured_text_bidi_override_options(Array p_args) {
 }
 
 Array Label::get_structured_text_bidi_override_options() const {
-	return st_args;
+	return Array(st_args);
 }
 
 Control::TextDirection Label::get_text_direction() const {
