@@ -34,7 +34,7 @@
 
 #include "modules/modules_enabled.gen.h"
 
-inline constexpr int default_theme_icons_count = 94;
+inline constexpr int default_theme_icons_count = 96;
 inline constexpr const char *default_theme_icons_sources[] = {
 	R"<!>(<svg xmlns="http://www.w3.org/2000/svg" width="16" height="16"><path fill="#b2b2b2" d="m7 1v6h-6v2h6v6h2v-6h6v-2h-6v-6z"/></svg>
 )<!>",
@@ -161,6 +161,10 @@ inline constexpr const char *default_theme_icons_sources[] = {
 	R"<!>(<svg xmlns="http://www.w3.org/2000/svg" width="16" height="16"><circle cx="8" cy="8" r="6" fill="#fefffe" fill-opacity=".75"/><path fill="none" stroke="#1a1a1a" stroke-linecap="round" stroke-linejoin="round" stroke-opacity=".65" stroke-width="2" d="m7 5 3 3-3 3"/></svg>
 )<!>",
 	R"<!>(<svg xmlns="http://www.w3.org/2000/svg" width="16" height="16"><circle cx="8" cy="8" r="6" fill="#fefffe"/><path fill="none" stroke="#1a1a1a" stroke-linecap="round" stroke-linejoin="round" stroke-width="2" d="m7 5 3 3-3 3"/></svg>
+)<!>",
+	R"<!>(<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="24" height="32" viewBox="0 0 6.35 8.467"><defs><linearGradient id="a"><stop offset="0" stop-color="#fff" stop-opacity=".3"/><stop offset="1" stop-color="#fff" stop-opacity="0"/></linearGradient><linearGradient xlink:href="#a" id="b" x1="4.233" x2="4.233" y1="0" y2="6.35" gradientTransform="translate(-8.467)" gradientUnits="userSpaceOnUse"/></defs><path fill="url(#b)" d="M-8.467 0H0V6.35H-8.467z" paint-order="fill markers stroke" transform="rotate(-90)"/></svg>
+)<!>",
+	R"<!>(<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="32" height="24" viewBox="0 0 8.467 6.35"><defs><linearGradient id="a"><stop offset="0" stop-color="#fff" stop-opacity=".3"/><stop offset="1" stop-color="#fff" stop-opacity="0"/></linearGradient><linearGradient xlink:href="#a" id="b" x1="4.233" x2="4.233" y1="0" y2="6.35" gradientUnits="userSpaceOnUse"/></defs><path fill="url(#b)" d="M0 0H8.467V6.35H0z" paint-order="fill markers stroke"/></svg>
 )<!>",
 	R"<!>(<svg xmlns="http://www.w3.org/2000/svg" width="16" height="16"><circle cx="8" cy="8" r="7" fill="#fefefe" fill-opacity=".75"/></svg>
 )<!>",
@@ -290,6 +294,8 @@ inline constexpr const char *default_theme_icons_names[] = {
 	"scroll_button_left_hl",
 	"scroll_button_right",
 	"scroll_button_right_hl",
+	"scroll_hint_horizontal",
+	"scroll_hint_vertical",
 	"slider_grabber",
 	"slider_grabber_disabled",
 	"slider_grabber_hl",
