@@ -79,21 +79,16 @@ void main() {
 
 layout(location = 0) in vec2 uv_interp;
 
-#ifdef SUBPASS
-layout(input_attachment_index = 0, set = 0, binding = 0) uniform subpassInput input_color;
-#elif defined(USE_MULTIVIEW)
-layout(set = 0, binding = 0) uniform sampler2DArray source_color;
+#ifdef USE_MULTIVIEW
+#define SAMPLER_FORMAT sampler2DArray
 #else
-layout(set = 0, binding = 0) uniform sampler2D source_color;
+#define SAMPLER_FORMAT sampler2D
 #endif
 
+layout(set = 0, binding = 0) uniform SAMPLER_FORMAT source_color;
 layout(set = 1, binding = 0) uniform sampler2D source_auto_exposure;
-#ifdef USE_MULTIVIEW
-layout(set = 2, binding = 0) uniform sampler2DArray source_glow;
-#else
-layout(set = 2, binding = 0) uniform sampler2D source_glow;
-#endif
-layout(set = 2, binding = 1) uniform sampler2D glow_map;
+layout(set = 2, binding = 0) uniform SAMPLER_FORMAT source_glow;
+layout(set = 2, binding = 1) uniform sampler2D glow_map; 
 
 #ifdef USE_1D_LUT
 layout(set = 3, binding = 0) uniform sampler2D source_color_correction;
@@ -107,8 +102,7 @@ layout(set = 3, binding = 0) uniform sampler3D source_color_correction;
 #define FLAG_USE_COLOR_CORRECTION (1 << 3)
 #define FLAG_USE_FXAA (1 << 4)
 #define FLAG_USE_8_BIT_DEBANDING (1 << 5)
-#define FLAG_USE_10_BIT_DEBANDING (1 << 6)
-#define FLAG_CONVERT_TO_SRGB (1 << 7)
+#define FLAG_CONVERT_TO_SRGB (1 << 6)
 
 layout(push_constant, std430) uniform Params {
 	vec3 bcs;
@@ -129,11 +123,192 @@ layout(push_constant, std430) uniform Params {
 	float white;
 	float auto_exposure_scale;
 	float luminance_multiplier;
+
+	vec4 tonemapper_params;
 }
 params;
 
 layout(location = 0) out vec4 frag_color;
 
+
+vec3 tonemap_reinhard(vec3 color) {
+	float white_squared = params.tonemapper_params.x;
+	vec3 white_squared_color = white_squared * color;
+	
+	return (white_squared_color + color * color) / (white_squared_color + white_squared);
+}
+
+vec3 tonemap_filmic(vec3 color) {
+	
+	
+	
+	
+	const float exposure_bias = 2.0f;
+	const float A = 0.22f * exposure_bias * exposure_bias; 
+	const float B = 0.30f * exposure_bias;
+	const float C = 0.10f;
+	const float D = 0.20f;
+	const float E = 0.01f;
+	const float F = 0.30f;
+
+	vec3 color_tonemapped = ((color * (A * color + C * B) + D * E) / (color * (A * color + B) + D * F)) - E / F;
+
+	return color_tonemapped / params.tonemapper_params.x;
+}
+
+
+
+vec3 tonemap_aces(vec3 color) {
+	
+	const float exposure_bias = 1.8f;
+	const float A = 0.0245786f;
+	const float B = 0.000090537f;
+	const float C = 0.983729f;
+	const float D = 0.432951f;
+	const float E = 0.238081f;
+
+	
+	const mat3 rgb_to_rrt = mat3(
+			vec3(0.59719f * exposure_bias, 0.35458f * exposure_bias, 0.04823f * exposure_bias),
+			vec3(0.07600f * exposure_bias, 0.90834f * exposure_bias, 0.01566f * exposure_bias),
+			vec3(0.02840f * exposure_bias, 0.13383f * exposure_bias, 0.83777f * exposure_bias));
+
+	const mat3 odt_to_rgb = mat3(
+			vec3(1.60475f, -0.53108f, -0.07367f),
+			vec3(-0.10208f, 1.10813f, -0.00605f),
+			vec3(-0.00327f, -0.07276f, 1.07602f));
+
+	color *= rgb_to_rrt;
+	vec3 color_tonemapped = (color * (color + A) - B) / (color * (C * color + D) + E);
+	color_tonemapped *= odt_to_rgb;
+
+	return color_tonemapped / params.tonemapper_params.x;
+}
+
+
+
+
+vec3 allenwp_curve(vec3 x) {
+	const float output_max_value = 1.0; 
+
+	
+	
+	const float awp_crossover_point = 0.18;
+	
+	
+	const float awp_shoulder_max = output_max_value - awp_crossover_point;
+
+	float awp_contrast = params.tonemapper_params.x;
+	float awp_toe_a = params.tonemapper_params.y;
+	float awp_slope = params.tonemapper_params.z;
+	float awp_w = params.tonemapper_params.w;
+
+	
+	vec3 s = x - awp_crossover_point;
+	vec3 slope_s = awp_slope * s;
+	s = slope_s * (1.0 + s / awp_w) / (1.0 + (slope_s / awp_shoulder_max));
+	s += awp_crossover_point;
+
+	
+	vec3 t = pow(x, vec3(awp_contrast));
+	t = t / (t + awp_toe_a);
+
+	return mix(s, t, lessThan(x, vec3(awp_crossover_point)));
+}
+
+
+
+
+
+vec3 tonemap_agx(vec3 color) {
+	
+	
+	
+	
+	
+	
+	
+	
+	
+	
+	
+	
+	
+	
+	
+	
+	
+
+	
+	const mat3 rec709_to_rec2020_agx_inset_matrix = mat3(
+			0.544814746488245, 0.140416948464053, 0.0888104196149096,
+			0.373787398372697, 0.754137554567394, 0.178871756420858,
+			0.0813978551390581, 0.105445496968552, 0.732317823964232);
+
+	
+	const mat3 agx_outset_rec2020_to_rec709_matrix = mat3(
+			1.96488741169489, -0.299313364904742, -0.164352742528393,
+			-0.855988495690215, 1.32639796461980, -0.238183969428088,
+			-0.108898916004672, -0.0270845997150571, 1.40253671195648);
+
+	const float output_max_value = 1.0; 
+
+	
+	color = rec709_to_rec2020_agx_inset_matrix * color;
+
+	
+	
+	color = allenwp_curve(color);
+
+	
+	
+	color = min(vec3(output_max_value), color);
+
+	
+	color = agx_outset_rec2020_to_rec709_matrix * color;
+
+	
+	
+	
+	return color;
+}
+
+vec3 linear_to_srgb(vec3 color) {
+	const vec3 a = vec3(0.055f);
+	return mix((vec3(1.0f) + a) * pow(color.rgb, vec3(1.0f / 2.4f)) - a, 12.92f * color.rgb, lessThan(color.rgb, vec3(0.0031308f)));
+}
+
+vec3 srgb_to_linear(vec3 color) {
+	const vec3 a = vec3(0.055f);
+	return mix(pow((color.rgb + a) * (1.0f / (vec3(1.0f) + a)), vec3(2.4f)), color.rgb * (1.0f / 12.92f), lessThan(color.rgb, vec3(0.04045f)));
+}
+
+#define TONEMAPPER_LINEAR 0
+#define TONEMAPPER_REINHARD 1
+#define TONEMAPPER_FILMIC 2
+#define TONEMAPPER_ACES 3
+#define TONEMAPPER_AGX 4
+
+vec3 apply_tonemapping(vec3 color) { 
+	if (params.tonemapper == TONEMAPPER_LINEAR) {
+		return color;
+	}
+
+	
+	
+	color = max(vec3(0.0), color);
+
+	if (params.tonemapper == TONEMAPPER_REINHARD) {
+		return tonemap_reinhard(color);
+	} else if (params.tonemapper == TONEMAPPER_FILMIC) {
+		return tonemap_filmic(color);
+	} else if (params.tonemapper == TONEMAPPER_ACES) {
+		return tonemap_aces(color);
+	} else { 
+		return tonemap_agx(color);
+	}
+}
+
 #ifdef USE_GLOW_FILTER_BICUBIC
 
 float w0(float a) {
@@ -239,171 +414,8 @@ vec4 texture2D_bicubic(sampler2D tex, vec2 uv, int p_lod) {
 
 #endif 
 
+vec3 gather_glow(SAMPLER_FORMAT tex, vec2 uv) { 
 
-vec3 tonemap_reinhard(vec3 color, float white) {
-	float white_squared = white * white;
-	vec3 white_squared_color = white_squared * color;
-	
-	return (white_squared_color + color * color) / (white_squared_color + white_squared);
-}
-
-vec3 tonemap_filmic(vec3 color, float white) {
-	
-	
-	
-	const float exposure_bias = 2.0f;
-	const float A = 0.22f * exposure_bias * exposure_bias; 
-	const float B = 0.30f * exposure_bias;
-	const float C = 0.10f;
-	const float D = 0.20f;
-	const float E = 0.01f;
-	const float F = 0.30f;
-
-	vec3 color_tonemapped = ((color * (A * color + C * B) + D * E) / (color * (A * color + B) + D * F)) - E / F;
-	float white_tonemapped = ((white * (A * white + C * B) + D * E) / (white * (A * white + B) + D * F)) - E / F;
-
-	return color_tonemapped / white_tonemapped;
-}
-
-
-
-vec3 tonemap_aces(vec3 color, float white) {
-	const float exposure_bias = 1.8f;
-	const float A = 0.0245786f;
-	const float B = 0.000090537f;
-	const float C = 0.983729f;
-	const float D = 0.432951f;
-	const float E = 0.238081f;
-
-	
-	const mat3 rgb_to_rrt = mat3(
-			vec3(0.59719f * exposure_bias, 0.35458f * exposure_bias, 0.04823f * exposure_bias),
-			vec3(0.07600f * exposure_bias, 0.90834f * exposure_bias, 0.01566f * exposure_bias),
-			vec3(0.02840f * exposure_bias, 0.13383f * exposure_bias, 0.83777f * exposure_bias));
-
-	const mat3 odt_to_rgb = mat3(
-			vec3(1.60475f, -0.53108f, -0.07367f),
-			vec3(-0.10208f, 1.10813f, -0.00605f),
-			vec3(-0.00327f, -0.07276f, 1.07602f));
-
-	color *= rgb_to_rrt;
-	vec3 color_tonemapped = (color * (color + A) - B) / (color * (C * color + D) + E);
-	color_tonemapped *= odt_to_rgb;
-
-	white *= exposure_bias;
-	float white_tonemapped = (white * (white + A) - B) / (white * (C * white + D) + E);
-
-	return color_tonemapped / white_tonemapped;
-}
-
-
-
-vec3 agx_contrast_approx(vec3 x) {
-	
-	
-	
-	
-	vec3 x2 = x * x;
-	vec3 x4 = x2 * x2;
-	return 0.021 * x + 4.0111 * x2 - 25.682 * x2 * x + 70.359 * x4 - 74.778 * x4 * x + 27.069 * x4 * x2;
-}
-
-
-
-
-vec3 tonemap_agx(vec3 color) {
-	
-	const mat3 srgb_to_rec2020_agx_inset_matrix = mat3(
-			0.54490813676363087053, 0.14044005884001287035, 0.088827411851915368603,
-			0.37377945959812267119, 0.75410959864013760045, 0.17887712465043811023,
-			0.081384976686407536266, 0.10543358536857773485, 0.73224999956948382528);
-
-	
-	const mat3 agx_outset_rec2020_to_srgb_matrix = mat3(
-			1.9645509602733325934, -0.29932243390911083839, -0.16436833806080403409,
-			-0.85585845117807513559, 1.3264510741502356555, -0.23822464068860595117,
-			-0.10886710826831608324, -0.027084020983874825605, 1.402665347143271889);
-
-	
-	
-	
-	const float min_ev = -12.4739311883324; 
-	const float max_ev = 4.02606881166759; 
-
-	
-	
-	
-	
-	
-	
-	
-	
-	
-	
-	
-	
-	
-	color = max(color, 2e-10);
-
-	
-	color = srgb_to_rec2020_agx_inset_matrix * color;
-
-	
-	
-	
-	color = clamp(log2(color), min_ev, max_ev);
-	color = (color - min_ev) / (max_ev - min_ev);
-
-	
-	color = agx_contrast_approx(color);
-
-	
-	color = pow(color, vec3(2.4));
-
-	
-	color = agx_outset_rec2020_to_srgb_matrix * color;
-
-	
-	
-	
-	return color;
-}
-
-vec3 linear_to_srgb(vec3 color) {
-	
-	
-	color = clamp(color, vec3(0.0), vec3(1.0));
-	const vec3 a = vec3(0.055f);
-	return mix((vec3(1.0f) + a) * pow(color.rgb, vec3(1.0f / 2.4f)) - a, 12.92f * color.rgb, lessThan(color.rgb, vec3(0.0031308f)));
-}
-
-#define TONEMAPPER_LINEAR 0
-#define TONEMAPPER_REINHARD 1
-#define TONEMAPPER_FILMIC 2
-#define TONEMAPPER_ACES 3
-#define TONEMAPPER_AGX 4
-
-vec3 apply_tonemapping(vec3 color, float white) { 
-	
-	
-	if (params.tonemapper == TONEMAPPER_LINEAR) {
-		return color;
-	} else if (params.tonemapper == TONEMAPPER_REINHARD) {
-		return tonemap_reinhard(max(vec3(0.0f), color), white);
-	} else if (params.tonemapper == TONEMAPPER_FILMIC) {
-		return tonemap_filmic(max(vec3(0.0f), color), white);
-	} else if (params.tonemapper == TONEMAPPER_ACES) {
-		return tonemap_aces(max(vec3(0.0f), color), white);
-	} else { 
-		return tonemap_agx(color);
-	}
-}
-
-#ifdef USE_MULTIVIEW
-vec3 gather_glow(sampler2DArray tex, vec2 uv) { 
-#else
-vec3 gather_glow(sampler2D tex, vec2 uv) { 
-#endif 
 	vec3 glow = vec3(0.0f);
 
 	if (params.glow_levels[0] > 0.0001) {
@@ -434,6 +446,8 @@ vec3 gather_glow(sampler2D tex, vec2 uv) {
 		glow += GLOW_TEXTURE_SAMPLE(tex, uv, 6).rgb * params.glow_levels[6];
 	}
 
+	glow = glow * params.luminance_multiplier;
+
 	return glow;
 }
 
@@ -443,34 +457,48 @@ vec3 gather_glow(sampler2D tex, vec2 uv) {
 #define GLOW_MODE_REPLACE 3
 #define GLOW_MODE_MIX 4
 
-vec3 apply_glow(vec3 color, vec3 glow) { 
+
+vec3 apply_glow(vec3 color, vec3 glow, float white) {
 	if (params.glow_mode == GLOW_MODE_ADD) {
 		return color + glow;
 	} else if (params.glow_mode == GLOW_MODE_SCREEN) {
 		
-		glow.rgb = clamp(glow.rgb, vec3(0.0f), vec3(1.0f));
-		return max((color + glow) - (color * glow), vec3(0.0));
+		
+		
+		
+		
+		glow.rgb = clamp(glow.rgb, 0.0, white);
+
+		
+		
+		
+		
+		
+		
+
+		
+		color.rgb = color.rgb + glow.rgb - (color.rgb * glow.rgb / white);
+
+		return color;
 	} else if (params.glow_mode == GLOW_MODE_SOFTLIGHT) {
 		
-		glow.rgb = clamp(glow.rgb, vec3(0.0f), vec3(1.0f));
-		glow = glow * vec3(0.5f) + vec3(0.5f);
+		
+		
+		
+		
+		
+		glow.rgb = clamp(glow.rgb, 0.0, 1.0);
+
+		color.r = color.r > 1.0 ? color.r : color.r + glow.r * ((color.r <= 0.25f ? ((16.0f * color.r - 12.0f) * color.r + 4.0f) * color.r : sqrt(color.r)) - color.r);
+		color.g = color.g > 1.0 ? color.g : color.g + glow.g * ((color.g <= 0.25f ? ((16.0f * color.g - 12.0f) * color.g + 4.0f) * color.g : sqrt(color.g)) - color.g);
+		color.b = color.b > 1.0 ? color.b : color.b + glow.b * ((color.b <= 0.25f ? ((16.0f * color.b - 12.0f) * color.b + 4.0f) * color.b : sqrt(color.b)) - color.b);
 
-		color.r = (glow.r <= 0.5f) ? (color.r - (1.0f - 2.0f * glow.r) * color.r * (1.0f - color.r)) : (((glow.r > 0.5f) && (color.r <= 0.25f)) ? (color.r + (2.0f * glow.r - 1.0f) * (4.0f * color.r * (4.0f * color.r + 1.0f) * (color.r - 1.0f) + 7.0f * color.r)) : (color.r + (2.0f * glow.r - 1.0f) * (sqrt(color.r) - color.r)));
-		color.g = (glow.g <= 0.5f) ? (color.g - (1.0f - 2.0f * glow.g) * color.g * (1.0f - color.g)) : (((glow.g > 0.5f) && (color.g <= 0.25f)) ? (color.g + (2.0f * glow.g - 1.0f) * (4.0f * color.g * (4.0f * color.g + 1.0f) * (color.g - 1.0f) + 7.0f * color.g)) : (color.g + (2.0f * glow.g - 1.0f) * (sqrt(color.g) - color.g)));
-		color.b = (glow.b <= 0.5f) ? (color.b - (1.0f - 2.0f * glow.b) * color.b * (1.0f - color.b)) : (((glow.b > 0.5f) && (color.b <= 0.25f)) ? (color.b + (2.0f * glow.b - 1.0f) * (4.0f * color.b * (4.0f * color.b + 1.0f) * (color.b - 1.0f) + 7.0f * color.b)) : (color.b + (2.0f * glow.b - 1.0f) * (sqrt(color.b) - color.b)));
 		return color;
 	} else { 
 		return glow;
 	}
 }
 
-vec3 apply_bcs(vec3 color, vec3 bcs) {
-	color = mix(vec3(0.0f), color, bcs.x);
-	color = mix(vec3(0.5f), color, bcs.y);
-	color = mix(vec3(dot(vec3(1.0f), color) * 0.33333f), color, bcs.z);
-
-	return color;
-}
 #ifdef USE_1D_LUT
 vec3 apply_color_correction(vec3 color) {
 	color.r = texture(source_color_correction, vec2(color.r, 0.0f)).r;
@@ -484,8 +512,6 @@ vec3 apply_color_correction(vec3 color) {
 }
 #endif
 
-#ifndef SUBPASS
-
 
 
 
@@ -644,7 +670,7 @@ vec3 do_fxaa(vec3 color, float exposure, vec2 uv_interp) {
 	if (!reached2) {
 		uv2 += vec3(offset * QUALITY(1), 0.0);
 	}
-)<!>" R"<!>(
+
 	if (!reachedBoth) {
 		for (int i = 2; i < ITERATIONS; i++) {
 			if (!reached1) {
@@ -688,7 +714,7 @@ vec3 do_fxaa(vec3 color, float exposure, vec2 uv_interp) {
 	float pixelOffset = -distanceFinal / edgeThickness + 0.5;
 
 	float finalOffset = correctVariation ? pixelOffset : 0.0;
-
+)<!>" R"<!>(
 	float lumaAverage = (1.0 / 12.0) * (2.0 * (lumaDownUp + lumaLeftRight) + lumaLeftCorners + lumaRightCorners);
 
 	float subPixelOffset1 = clamp(abs(lumaAverage - lumaCenter) / lumaRange, 0.0, 1.0);
@@ -854,7 +880,6 @@ vec3 do_fxaa(vec3 color, float exposure, vec2 uv_interp) {
 
 #endif
 }
-#endif 
 
 
 
@@ -873,15 +898,7 @@ vec3 screen_space_dither(vec2 frag_coord, float bit_alignment_diviser) {
 }
 
 void main() {
-#ifdef SUBPASS
-	
 #ifdef USE_MULTIVIEW
-	
-	
-	uint vi = ViewIndex;
-#endif
-	vec4 color = subpassLoad(input_color);
-#elif defined(USE_MULTIVIEW)
 	vec4 color = textureLod(source_color, vec3(uv_interp, ViewIndex), 0.0f);
 #else
 	vec4 color = textureLod(source_color, uv_interp, 0.0f);
@@ -892,68 +909,84 @@ void main() {
 
 	float exposure = params.exposure;
 
-#ifndef SUBPASS
 	if (bool(params.flags & FLAG_USE_AUTO_EXPOSURE)) {
 		exposure *= 1.0 / (texelFetch(source_auto_exposure, ivec2(0, 0), 0).r * params.luminance_multiplier / params.auto_exposure_scale);
 	}
-#endif
 
 	color.rgb *= exposure;
 
 	
-#ifndef SUBPASS
 	if (bool(params.flags & FLAG_USE_FXAA)) {
 		
 		color.rgb = do_fxaa(color.rgb, exposure, uv_interp);
 	}
 
-	if (bool(params.flags & FLAG_USE_GLOW) && params.glow_mode == GLOW_MODE_MIX) {
-		vec3 glow = gather_glow(source_glow, uv_interp) * params.luminance_multiplier;
+	if (bool(params.flags & FLAG_USE_GLOW) && params.glow_mode != GLOW_MODE_SOFTLIGHT) {
+		vec3 glow = gather_glow(source_glow, uv_interp) * params.glow_intensity;
 		if (params.glow_map_strength > 0.001) {
 			glow = mix(glow, texture(glow_map, uv_interp).rgb * glow, params.glow_map_strength);
 		}
-		color.rgb = mix(color.rgb, glow, params.glow_intensity);
-	}
-#endif
 
-	color.rgb = apply_tonemapping(color.rgb, params.white);
-
-	bool convert_to_srgb = bool(params.flags & FLAG_CONVERT_TO_SRGB);
-	if (convert_to_srgb) {
-		color.rgb = linear_to_srgb(color.rgb); 
+		if (params.glow_mode == GLOW_MODE_MIX) {
+			color.rgb = color.rgb * (1.0 - params.glow_intensity) + glow;
+		} else {
+			color.rgb = apply_glow(color.rgb, glow, params.white);
+		}
 	}
-#ifndef SUBPASS
+
 	
-	if (bool(params.flags & FLAG_USE_GLOW) && params.glow_mode != GLOW_MODE_MIX) {
-		vec3 glow = gather_glow(source_glow, uv_interp) * params.glow_intensity * params.luminance_multiplier;
+
+	color.rgb = apply_tonemapping(color.rgb);
+
+	
+
+	if (bool(params.flags & FLAG_USE_GLOW) && params.glow_mode == GLOW_MODE_SOFTLIGHT) {
+		
+		
+		
+		vec3 glow = gather_glow(source_glow, uv_interp) * params.glow_intensity;
 		if (params.glow_map_strength > 0.001) {
 			glow = mix(glow, texture(glow_map, uv_interp).rgb * glow, params.glow_map_strength);
 		}
-
-		
-		glow = apply_tonemapping(glow, params.white);
-		if (convert_to_srgb) {
-			glow = linear_to_srgb(glow);
-		}
-
-		color.rgb = apply_glow(color.rgb, glow);
+		glow = apply_tonemapping(glow);
+		color.rgb = apply_glow(color.rgb, glow, params.white);
 	}
-#endif
 
 	
 
 	if (bool(params.flags & FLAG_USE_BCS)) {
-		color.rgb = apply_bcs(color.rgb, params.bcs);
-	}
+		
+		
+		
+		
+		color.rgb = color.rgb * params.bcs.x;
+
+		color.rgb = linear_to_srgb(color.rgb);
 
-	if (bool(params.flags & FLAG_USE_COLOR_CORRECTION)) {
 		
-		if (!convert_to_srgb) {
-			color.rgb = linear_to_srgb(color.rgb);
+		
+		
+		
+		
+		color.rgb = mix(vec3(0.5), color.rgb, params.bcs.y);
+
+		
+		
+		
+		
+		color.rgb = mix(vec3(dot(vec3(1.0), color.rgb) * (1.0 / 3.0)), color.rgb, params.bcs.z);
+
+		if (bool(params.flags & FLAG_USE_COLOR_CORRECTION)) {
+			color.rgb = clamp(color.rgb, vec3(0.0), vec3(1.0));
+			color.rgb = apply_color_correction(color.rgb);
+			
+			
+			
+		} else if (!bool(params.flags & FLAG_CONVERT_TO_SRGB)) {
+			color.rgb = srgb_to_linear(color.rgb);
 		}
-		color.rgb = apply_color_correction(color.rgb);
-		
-		
+	} else if (bool(params.flags & FLAG_CONVERT_TO_SRGB)) {
+		color.rgb = linear_to_srgb(color.rgb);
 	}
 
 	
@@ -962,9 +995,6 @@ void main() {
 	if (bool(params.flags & FLAG_USE_8_BIT_DEBANDING)) {
 		
 		color.rgb += screen_space_dither(gl_FragCoord.xy, 255.0);
-	} else if (bool(params.flags & FLAG_USE_10_BIT_DEBANDING)) {
-		
-		color.rgb += screen_space_dither(gl_FragCoord.xy, 1023.0);
 	}
 
 	frag_color = color;
