@@ -45,6 +45,58 @@ public:
 
 #VERSION_DEFINES
 
+
+vec3 oct_to_vec3(vec2 e) {
+	vec3 v = vec3(e.xy, 1.0 - abs(e.x) - abs(e.y));
+	float t = max(-v.z, 0.0);
+	v.xy += t * -sign(v.xy);
+	return normalize(v);
+}
+
+
+vec3 oct_to_vec3_with_border(vec2 uv, float border_size) {
+	
+	uv = (uv - 0.5) * (2.0 / border_size);
+	
+	
+	vec2 mask = step(vec2(1.0), abs(uv));
+	uv = 2.0 * clamp(uv, -1.0, 1.0) - uv;
+	uv = mix(uv, -uv, mask.yx);
+	return oct_to_vec3(uv);
+}
+
+vec2 oct_wrap(vec2 v) {
+	vec2 signVal;
+	signVal.x = v.x >= 0.0 ? 1.0 : -1.0;
+	signVal.y = v.y >= 0.0 ? 1.0 : -1.0;
+	return (1.0 - abs(v.yx)) * signVal;
+}
+
+vec2 vec3_to_oct(vec3 n) {
+	
+	n /= (abs(n.x) + abs(n.y) + abs(n.z));
+	n.xy = (n.z >= 0.0) ? n.xy : oct_wrap(n.xy);
+	n.xy = n.xy * 0.5 + 0.5;
+	return n.xy;
+}
+
+
+
+vec2 vec3_to_oct_with_border(vec3 n, vec2 border_size) {
+	vec2 uv = vec3_to_oct(n);
+	return uv * border_size.y + border_size.x;
+}
+
+float vec3_to_oct_lod(vec3 n_ddx, vec3 n_ddy, float pixel_size) {
+	
+	
+	float pixel_size_sqr = 4.0 * pixel_size * pixel_size;
+	float ddx = dot(n_ddx, n_ddx) / pixel_size_sqr;
+	float ddy = dot(n_ddy, n_ddy) / pixel_size_sqr;
+	float dd_sqr = max(ddx, ddy);
+	return 0.25 * log2(dd_sqr + 1e-6f);
+}
+
 layout(local_size_x = 8, local_size_y = 8, local_size_z = 1) in;
 
 #define MAX_CASCADES 8
@@ -81,10 +133,10 @@ layout(rgba32i, set = 0, binding = 13) uniform restrict iimage2D lightprobe_aver
 
 layout(rgba16f, set = 0, binding = 14) uniform restrict writeonly image2DArray lightprobe_ambient_texture;
 
-#ifdef USE_CUBEMAP_ARRAY
-layout(set = 1, binding = 0) uniform textureCubeArray sky_irradiance;
+#ifdef USE_RADIANCE_OCTMAP_ARRAY
+layout(set = 1, binding = 0) uniform texture2DArray sky_irradiance;
 #else
-layout(set = 1, binding = 0) uniform textureCube sky_irradiance;
+layout(set = 1, binding = 0) uniform texture2D sky_irradiance;
 #endif
 layout(set = 1, binding = 1) uniform sampler linear_sampler_mipmaps;
 
@@ -116,8 +168,9 @@ layout(push_constant, std430) uniform Params {
 	vec3 sky_color_or_orientation;
 	float y_mult;
 
+	vec2 sky_irradiance_border_size;
 	bool store_ambient_texture;
-	uint pad[3];
+	uint pad;
 }
 params;
 
@@ -312,10 +365,10 @@ void main() {
 			vec4 sky_quat = vec4(params.sky_color_or_orientation, sky_sign * sqrt(1.0 - dot(params.sky_color_or_orientation, params.sky_color_or_orientation)));
 			vec3 sky_dir = cross(sky_quat.xyz, ray_dir);
 			sky_dir = ray_dir + ((sky_dir * sky_quat.w) + cross(sky_quat.xyz, sky_dir)) * 2.0;
-#ifdef USE_CUBEMAP_ARRAY
-			light.rgb = textureLod(samplerCubeArray(sky_irradiance, linear_sampler_mipmaps), vec4(sky_dir, 0.0), 2.0).rgb; 
+#ifdef USE_RADIANCE_OCTMAP_ARRAY
+			light.rgb = textureLod(sampler2DArray(sky_irradiance, linear_sampler_mipmaps), vec3(vec3_to_oct_with_border(sky_dir, params.sky_irradiance_border_size), 0.0), 2.0).rgb; 
 #else
-			light.rgb = textureLod(samplerCube(sky_irradiance, linear_sampler_mipmaps), sky_dir, 2.0).rgb; 
+			light.rgb = textureLod(sampler2D(sky_irradiance, linear_sampler_mipmaps), vec3_to_oct_with_border(sky_dir, params.sky_irradiance_border_size), 2.0).rgb; 
 #endif
 			light.rgb *= params.sky_energy;
 			light.a = 0.0;
@@ -524,7 +577,7 @@ void main() {
 		ivec2 tex_pos;
 		tex_pos = read_probe.xy;
 		tex_pos.x += read_probe.z * int(params.probe_axis_size);
-
+)<!>" R"<!>(
 		
 		for (uint j = 0; j < params.history_size; j++) {
 			for (int i = 0; i < SH_SIZE; i++) {
@@ -559,7 +612,7 @@ void main() {
 
 		ivec3 probe_posi = ivec3(probe_pos);
 		
-)<!>" R"<!>(
+
 		vec4 average_light[SH_SIZE] = vec4[](vec4(0), vec4(0), vec4(0), vec4(0), vec4(0), vec4(0), vec4(0), vec4(0), vec4(0)
 #if (SH_SIZE == 16)
 																															 ,
