@@ -146,6 +146,57 @@ struct DirectionalLightData {
 	vec2 uv_scale4;
 };
 
+vec3 oct_to_vec3(vec2 e) {
+	vec3 v = vec3(e.xy, 1.0 - abs(e.x) - abs(e.y));
+	float t = max(-v.z, 0.0);
+	v.xy += t * -sign(v.xy);
+	return normalize(v);
+}
+
+
+vec3 oct_to_vec3_with_border(vec2 uv, float border_size) {
+	
+	uv = (uv - 0.5) * (2.0 / border_size);
+	
+	
+	vec2 mask = step(vec2(1.0), abs(uv));
+	uv = 2.0 * clamp(uv, -1.0, 1.0) - uv;
+	uv = mix(uv, -uv, mask.yx);
+	return oct_to_vec3(uv);
+}
+
+vec2 oct_wrap(vec2 v) {
+	vec2 signVal;
+	signVal.x = v.x >= 0.0 ? 1.0 : -1.0;
+	signVal.y = v.y >= 0.0 ? 1.0 : -1.0;
+	return (1.0 - abs(v.yx)) * signVal;
+}
+
+vec2 vec3_to_oct(vec3 n) {
+	
+	n /= (abs(n.x) + abs(n.y) + abs(n.z));
+	n.xy = (n.z >= 0.0) ? n.xy : oct_wrap(n.xy);
+	n.xy = n.xy * 0.5 + 0.5;
+	return n.xy;
+}
+
+
+
+vec2 vec3_to_oct_with_border(vec3 n, vec2 border_size) {
+	vec2 uv = vec3_to_oct(n);
+	return uv * border_size.y + border_size.x;
+}
+
+float vec3_to_oct_lod(vec3 n_ddx, vec3 n_ddy, float pixel_size) {
+	
+	
+	float pixel_size_sqr = 4.0 * pixel_size * pixel_size;
+	float ddx = dot(n_ddx, n_ddx) / pixel_size_sqr;
+	float ddy = dot(n_ddy, n_ddy) / pixel_size_sqr;
+	float dd_sqr = max(ddx, ddy);
+	return 0.25 * log2(dd_sqr + 1e-6f);
+}
+
 #define M_PI 3.14159265359
 
 #define DENSITY_SCALE 1024.0
@@ -305,6 +356,9 @@ layout(set = 0, binding = 14, std140) uniform Params {
 	uint temporal_frame;
 	float temporal_blend;
 
+	vec2 sky_border_size;
+	vec2 pad;
+
 	mat3x4 cam_rotation;
 	mat4 to_prev_view;
 
@@ -330,10 +384,10 @@ layout(r32ui, set = 0, binding = 17) uniform uimage3D light_only_map;
 layout(r32ui, set = 0, binding = 18) uniform uimage3D emissive_only_map;
 #endif
 
-#ifdef USE_RADIANCE_CUBEMAP_ARRAY
-layout(set = 0, binding = 19) uniform textureCubeArray sky_texture;
+#ifdef USE_RADIANCE_OCTMAP_ARRAY
+layout(set = 0, binding = 19) uniform texture2DArray sky_texture;
 #else
-layout(set = 0, binding = 19) uniform textureCube sky_texture;
+layout(set = 0, binding = 19) uniform texture2D sky_texture;
 #endif
 #endif 
 
@@ -558,12 +612,12 @@ void main() {
 			if (params.sky_contribution > 0.0) {
 				float mip_bias = 2.0 + total_density * (MAX_SKY_LOD - 2.0); 
 				vec3 scatter_direction = (params.radiance_inverse_xform * normalize(view_pos)) * sign(params.phase_g);
-#ifdef USE_RADIANCE_CUBEMAP_ARRAY
-				isotropic = texture(samplerCubeArray(sky_texture, linear_sampler_with_mipmaps), vec4(0.0, 1.0, 0.0, mip_bias)).rgb;
-				anisotropic = texture(samplerCubeArray(sky_texture, linear_sampler_with_mipmaps), vec4(scatter_direction, mip_bias)).rgb;
+#ifdef USE_RADIANCE_OCTMAP_ARRAY
+				isotropic = texture(sampler2DArray(sky_texture, linear_sampler_with_mipmaps), vec3(vec3_to_oct_with_border(vec3(0.0, 1.0, 0.0), params.sky_border_size), mip_bias)).rgb;
+				anisotropic = texture(sampler2DArray(sky_texture, linear_sampler_with_mipmaps), vec3(vec3_to_oct_with_border(scatter_direction, params.sky_border_size), mip_bias)).rgb;
 #else
-				isotropic = textureLod(samplerCube(sky_texture, linear_sampler_with_mipmaps), vec3(0.0, 1.0, 0.0), mip_bias).rgb;
-				anisotropic = textureLod(samplerCube(sky_texture, linear_sampler_with_mipmaps), vec3(scatter_direction), mip_bias).rgb;
+				isotropic = textureLod(sampler2D(sky_texture, linear_sampler_with_mipmaps), vec3_to_oct_with_border(vec3(0.0, 1.0, 0.0), params.sky_border_size), mip_bias).rgb;
+				anisotropic = textureLod(sampler2D(sky_texture, linear_sampler_with_mipmaps), vec3_to_oct_with_border(scatter_direction, params.sky_border_size), mip_bias).rgb;
 #endif 
 			}
 
@@ -580,7 +634,7 @@ void main() {
 			uint item_max;
 			uint item_from;
 			uint item_to;
-
+)<!>" R"<!>(
 			cluster_get_item_range(cluster_omni_offset + params.max_cluster_element_count_div_32 + cluster_z, item_min, item_max, item_from, item_to);
 
 			for (uint i = item_from; i < item_to; i++) {
@@ -613,7 +667,7 @@ void main() {
 							vec2 flip_offset = omni_lights.data[light_index].direction.xy;
 
 							vec3 local_vert = (omni_lights.data[light_index].shadow_matrix * vec4(view_pos, 1.0)).xyz;
-)<!>" R"<!>(
+
 							float shadow_len = length(local_vert); 
 							vec3 shadow_sample = normalize(local_vert);
 
@@ -688,7 +742,7 @@ void main() {
 							vec4 v = vec4(view_pos, 1.0);
 
 							vec4 splane = (spot_lights.data[light_index].shadow_matrix * v);
-							splane.z -= spot_lights.data[light_index].shadow_bias / (d * spot_lights.data[light_index].inv_radius);
+							splane.z -= spot_lights.data[light_index].shadow_bias;
 							splane /= splane.w;
 
 							vec3 pos = vec3(splane.xy * spot_lights.data[light_index].atlas_rect.zw + spot_lights.data[light_index].atlas_rect.xy, splane.z);
@@ -848,7 +902,7 @@ void main() {
 
 		prev_z = z;
 
-		imageStore(fog_map, fog_pos, vec4(fog_accum.rgb, 1.0 - fog_accum.a));
+		imageStore(fog_map, fog_pos, vec4(fog_accum.rgb, fog_accum.a));
 	}
 
 #endif
